"""
Based on:
https://msal-python.readthedocs.io/en/latest/#msal.SerializableTokenCache
"""
import os
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Union

from appdirs import user_cache_dir
from msal import SerializableTokenCache


class _BaseTokenCache(ABC, SerializableTokenCache):
    """
    Base class for a token cache
    """

    @abstractmethod
    def write_cache(self) -> None:
        """
        Write cache if needed.
        """
        raise NotImplementedError

    def __del__(self):
        self.write_cache()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.write_cache()


class SimpleTokenCache(_BaseTokenCache):
    """
    Provides a simple token cache for users to
    persist the cache across sessions.

    .. versionadded:: 0.4.0

    """

    def __init__(self, cache_file: Union[str, os.PathLike, None] = None) -> None:
        """
        Parameters
        ----------
        cache_file: Union[str, os.PathLike, None], optional
            Path to the token cache fike. If not provided,
            it will store one for you in the user cache directory.
        """
        super().__init__()
        if cache_file is None:
            self.cache_file = Path(
                user_cache_dir("msal-requests-auth", appauthor=False), "token-cache.bin"
            )
            self.cache_file.parent.mkdir(parents=True, exist_ok=True)
        else:
            self.cache_file = Path(cache_file)

        if self.cache_file.exists():
            self.deserialize(self.cache_file.read_text())

    def write_cache(self) -> None:
        """
        Write cache to disk if needed.
        """
        if self.has_state_changed:
            self.cache_file.write_text(self.serialize())


def _import_keyring():
    """
    Method to import keyring with error message
    """
    try:
        import keyring
    except ModuleNotFoundError as error:
        raise ModuleNotFoundError(
            "Please install msal_requests_auth with the "
            "'keyring' extra: msal_requests_auth[keyring]."
        ) from error
    return keyring


class KeyringTokenCache(_BaseTokenCache):
    """
    Provides a token cache for users to
    persist the cache across sessions using keyring.

    .. versionadded:: 0.7.0

    .. note:: Requires keyring to be installed. The 'keyring'
              extra can be used for that (msal_requests_auth[keyring]).
    """

    def __init__(self) -> None:
        super().__init__()
        token_cache = _import_keyring().get_password("__msal_requests_auth__", "token")
        if token_cache is not None:
            self.deserialize(token_cache)

    def write_cache(self) -> None:
        """
        Write cache to keyring if needed.
        """
        if self.has_state_changed:
            _import_keyring().set_password(
                "__msal_requests_auth__", "token", self.serialize()
            )



from unittest.mock import patch

from msal_requests_auth.cache import KeyringTokenCache, SimpleTokenCache


@patch("msal_requests_auth.cache.user_cache_dir")
@patch("msal_requests_auth.cache.SimpleTokenCache.serialize")
def test_simple_token_cache(serialize_mock, user_cache_dir_mock, tmp_path):
    serialize_mock.return_value = "TEST"
    user_cache_dir_mock.return_value = str(tmp_path / "msal-requests-auth")
    with SimpleTokenCache() as cache:
        cache.has_state_changed = True
    assert cache.cache_file.read_text() == "TEST"
    user_cache_dir_mock.assert_called_with("msal-requests-auth", appauthor=False)
    assert cache.cache_file == tmp_path / "msal-requests-auth" / "token-cache.bin"
    assert (tmp_path / "msal-requests-auth").exists()


@patch("msal_requests_auth.cache.user_cache_dir")
@patch("msal_requests_auth.cache.SimpleTokenCache.serialize")
def test_simple_token_cache__write_cache(serialize_mock, user_cache_dir_mock, tmp_path):
    serialize_mock.return_value = "TEST"
    user_cache_dir_mock.return_value = str(tmp_path / "msal-requests-auth")
    cache = SimpleTokenCache()
    cache.write_cache()
    assert not cache.cache_file.exists()
    cache.has_state_changed = True
    cache.write_cache()
    assert cache.cache_file.read_text() == "TEST"


@patch("msal_requests_auth.cache.user_cache_dir")
def test_simple_token_cache__custom_path(user_cache_dir_mock, tmp_path):
    test_file = tmp_path / "test.bin"
    cache = SimpleTokenCache(test_file)
    user_cache_dir_mock.assert_not_called()
    assert cache.cache_file == test_file


@patch("msal_requests_auth.cache.user_cache_dir")
@patch("msal_requests_auth.cache.SimpleTokenCache.deserialize")
def test_simple_token_cache__deserialize(
    deserialize_mock, user_cache_dir_mock, tmp_path
):
    test_file = tmp_path / "test.bin"
    test_file.write_text("TEST")
    cache = SimpleTokenCache(test_file)
    deserialize_mock.assert_called_with("TEST")
    user_cache_dir_mock.assert_not_called()
    assert cache.cache_file == test_file


@patch("msal_requests_auth.cache._import_keyring")
@patch("msal_requests_auth.cache.KeyringTokenCache.serialize")
def test_keyring_token_cache(serialize_mock, keyring_mock):
    serialize_mock.return_value = "TEST"
    keyring_mock.return_value.get_password.return_value = None
    with KeyringTokenCache() as cache:
        cache.has_state_changed = True
    keyring_mock.return_value.get_password.assert_called_with(
        "__msal_requests_auth__", "token"
    )
    keyring_mock.return_value.set_password.assert_called_with(
        "__msal_requests_auth__", "token", "TEST"
    )


@patch("msal_requests_auth.cache._import_keyring")
@patch("msal_requests_auth.cache.KeyringTokenCache.serialize")
def test_keyring_token_cache__write_cache(serialize_mock, keyring_mock):
    serialize_mock.return_value = "TEST"
    keyring_mock.return_value.get_password.return_value = None
    cache = KeyringTokenCache()
    keyring_mock.return_value.get_password.assert_called_with(
        "__msal_requests_auth__", "token"
    )
    cache.write_cache()
    keyring_mock.return_value.set_password.assert_not_called()
    cache.has_state_changed = True
    cache.write_cache()
    keyring_mock.return_value.set_password.assert_called_with(
        "__msal_requests_auth__", "token", "TEST"
    )


@patch("msal_requests_auth.cache._import_keyring")
@patch("msal_requests_auth.cache.KeyringTokenCache.deserialize")
def test_keyring_token_cache__deserialize(deserialize_mock, keyring_mock):
    keyring_mock.return_value.get_password.return_value = "TEST"
    KeyringTokenCache()
    deserialize_mock.assert_called_with("TEST")



"""
Module for handling the Device Code flow with MSAL and credential refresh.
"""
from typing import Dict

from msal import ConfidentialClientApplication

from .base_auth_client import BaseMSALRefreshAuth


class ClientCredentialAuth(BaseMSALRefreshAuth):
    """
    Auth class for the client credential flow with MSAL
    """

    _client_class = ConfidentialClientApplication

    def _get_access_token(self) -> Dict[str, str]:
        """
        Retrieve access token from MSAL using client credential flow.

        Based on: https://docs.microsoft.com/en-us/azure/active-directory/develop/scenario-daemon-acquire-token?tabs=python#acquiretokenforclient-api
        """
        result = self.client.acquire_token_silent(scopes=self.scopes, account=None)
        if not result:
            # "No suitable token exists in cache. Get a new one from AAD
            result = self.client.acquire_token_for_client(scopes=self.scopes)
        return result



from unittest.mock import MagicMock, patch

import pytest

from msal_requests_auth.auth import ClientCredentialAuth
from msal_requests_auth.exceptions import AuthenticationError


@patch("msal.ConfidentialClientApplication", autospec=True)
def test_client_credential_auth__no_cache(cca_mock):
    cca_mock.acquire_token_silent.return_value = None
    cca_mock.acquire_token_for_client.return_value = {
        "token_type": "Bearer",
        "access_token": "TEST TOKEN",
    }
    request_mock = MagicMock()
    request_mock.headers = {}
    returned_request = ClientCredentialAuth(client=cca_mock, scopes=["TEST SCOPE"])(
        request_mock
    )
    cca_mock.acquire_token_silent.assert_called_with(
        scopes=["TEST SCOPE"], account=None
    )
    cca_mock.acquire_token_for_client.assert_called_with(scopes=["TEST SCOPE"])

    assert returned_request.headers == {"Authorization": "Bearer TEST TOKEN"}


@patch("msal.ConfidentialClientApplication", autospec=True)
def test_client_credential_auth__unable_to_get_token(cca_mock):
    cca_mock.acquire_token_silent.return_value = None
    cca_mock.acquire_token_for_client.return_value = {
        "error": "BAD REQUEST",
        "error_description": "Request to get token was bad.",
    }
    request_mock = MagicMock()
    request_mock.headers = {}
    with pytest.raises(
        AuthenticationError,
        match=(
            r"Unable to get token\. Error: BAD REQUEST "
            r"\(Details: Request to get token was bad\.\)\."
        ),
    ):
        ClientCredentialAuth(client=cca_mock, scopes=["TEST SCOPE"])(request_mock)

    cca_mock.acquire_token_silent.assert_called_with(
        scopes=["TEST SCOPE"], account=None
    )
    cca_mock.acquire_token_for_client.assert_called_with(scopes=["TEST SCOPE"])


@patch("msal.ConfidentialClientApplication", autospec=True)
def test_client_credential_auth__cache(cca_mock):
    cca_mock.acquire_token_silent.return_value = {
        "token_type": "Bearer",
        "access_token": "TEST TOKEN",
    }
    request_mock = MagicMock()
    request_mock.headers = {}
    returned_request = ClientCredentialAuth(client=cca_mock, scopes=["TEST SCOPE"])(
        request_mock
    )
    cca_mock.acquire_token_silent.assert_called_with(
        scopes=["TEST SCOPE"], account=None
    )
    cca_mock.acquire_token_for_client.assert_not_called()

    assert returned_request.headers == {"Authorization": "Bearer TEST TOKEN"}



"""
Handles refresing tokens with MSAL.
"""
from abc import abstractmethod
from typing import Dict, List, Type

import msal
import requests

from msal_requests_auth.exceptions import AuthenticationError


class BaseMSALRefreshAuth(requests.auth.AuthBase):
    """
    Auth class for the device code flow with MSAL
    """

    def __init__(self, client: msal.ClientApplication, scopes: List[str]):
        """
        Parameters
        ----------
        client: msal.ClientApplication
            The MSAL client to use to get tokens.
        scopes: List[str]
            List of scopes to get token for.
        """
        if not isinstance(client, self._client_class):
            raise ValueError(
                f"Invalid client provided. Must be a {self._client_class.__name__}."
            )
        self.client = client
        self.scopes = scopes

    @property
    @abstractmethod
    def _client_class(self) -> Type[msal.ClientApplication]:
        """
        This is the expected type of the client class.
        """
        raise NotImplementedError

    def __call__(
        self, input_request: requests.PreparedRequest
    ) -> requests.PreparedRequest:
        """
        Adds the token to the authorization header.
        """
        token = self._get_access_token()
        if "access_token" not in token:
            error = token.get("error")
            description = token.get("error_description")
            raise AuthenticationError(
                f"Unable to get token. Error: {error} (Details: {description})."
            )
        input_request.headers[
            "Authorization"
        ] = f"{token['token_type']} {token['access_token']}"
        return input_request

    @abstractmethod
    def _get_access_token(self) -> Dict[str, str]:
        """
        Retrieves the token dictionary from Azure AD.

        Returns
        -------
        dict
        """
        raise NotImplementedError



"""
Module for handling the Device Code flow with MSAL and credential refresh.
"""
import os
import webbrowser
from typing import Dict, List, Optional

import pyperclip
from msal import PublicClientApplication

from .base_auth_client import BaseMSALRefreshAuth


class DeviceCodeAuth(BaseMSALRefreshAuth):
    """
    Auth class for the device code flow with MSAL
    """

    _client_class = PublicClientApplication

    def __init__(
        self,
        client: PublicClientApplication,
        scopes: List[str],
        headless: Optional[bool] = None,
    ):
        """
        .. versionadded:: 0.2.0 headless
        .. versionadded:: 0.6.0 MSAL_REQUESTS_AUTH_HEADLESS environment variable

        Parameters
        ----------
        client: msal.PublicClientApplication
            The MSAL client to use to get tokens.
        scopes: List[str]
            List of scopes to get token for.
        headless: bool, optional
            If None (default), it will check the MSAL_REQUESTS_AUTH_HEADLESS environment
            variable and default to False if it is not found.
            If False, it will open a webbrowser and copy the code to the clipboard.
            If True, it will skip automatically opening webbrowser and copying to clipboard.
        """
        super().__init__(client, scopes)
        headless_default = bool(os.getenv("MSAL_REQUESTS_AUTH_HEADLESS", False))
        self._headless = headless_default if headless is None else headless

    def _get_access_token(self) -> Dict[str, str]:
        """
        Retrieve access token from MSAL using device code flow.

        Based on README: https://github.com/AzureAD/microsoft-authentication-library-for-python
        """
        accounts = self.client.get_accounts()
        result = None
        if accounts:
            # use MSAL cache if available
            result = self.client.acquire_token_silent(
                scopes=self.scopes,
                account=accounts[0],
            )
            if result:
                return result
        # "No suitable token exists in cache. Get a new one from AAD
        flow = self.client.initiate_device_flow(
            scopes=self.scopes,
        )
        print(flow["message"])
        if not self._headless:
            # copy code to clipboard
            pyperclip.copy(flow["user_code"])
            webbrowser.open(flow["verification_uri"])
        return self.client.acquire_token_by_device_flow(flow)



import os
from unittest.mock import MagicMock, patch

import pytest

from msal_requests_auth.auth import DeviceCodeAuth
from msal_requests_auth.exceptions import AuthenticationError


@patch.dict(os.environ, {}, clear=True)
@pytest.mark.parametrize("headless", [True, False])
@patch("msal.PublicClientApplication", autospec=True)
@patch("msal_requests_auth.auth.device_code.webbrowser")
@patch("msal_requests_auth.auth.device_code.pyperclip")
def test_device_code_auth__no_accounts(
    pyperclip_patch, webbrowser_patch, pca_mock, headless
):
    pca_mock.get_accounts.return_value = None
    pca_mock.initiate_device_flow.return_value = {
        "message": "TEST MESSAGE",
        "verification_uri": "TEST URL",
        "user_code": "TEST CODE",
    }
    pca_mock.acquire_token_by_device_flow.return_value = {
        "token_type": "Bearer",
        "access_token": "TEST TOKEN",
    }
    request_mock = MagicMock()
    request_mock.headers = {}
    returned_request = DeviceCodeAuth(
        client=pca_mock, scopes=["TEST SCOPE"], headless=headless
    )(request_mock)
    pca_mock.acquire_token_silent.assert_not_called()
    pca_mock.initiate_device_flow.assert_called_with(scopes=["TEST SCOPE"])
    pca_mock.acquire_token_by_device_flow.assert_called_with(
        {
            "message": "TEST MESSAGE",
            "verification_uri": "TEST URL",
            "user_code": "TEST CODE",
        }
    )

    assert returned_request.headers == {"Authorization": "Bearer TEST TOKEN"}
    if headless:
        webbrowser_patch.open.assert_not_called()
        pyperclip_patch.copy.assert_not_called()
    else:
        pyperclip_patch.copy.assert_called_with("TEST CODE")
        webbrowser_patch.open.assert_called_with("TEST URL")


@patch.dict(os.environ, {"MSAL_REQUESTS_AUTH_HEADLESS": "true"}, clear=True)
@pytest.mark.parametrize("headless", [True, False, None])
@patch("msal.PublicClientApplication", autospec=True)
def test_device_code_auth__headless(pca_mock, headless):
    returned_request = DeviceCodeAuth(
        client=pca_mock, scopes=["TEST SCOPE"], headless=headless
    )
    if headless is False:
        assert returned_request._headless is False
    else:
        assert returned_request._headless is True


@patch.dict(os.environ, {}, clear=True)
@patch("msal.PublicClientApplication", autospec=True)
@patch("msal_requests_auth.auth.device_code.webbrowser")
@patch("msal_requests_auth.auth.device_code.pyperclip")
def test_device_code_auth__no_accounts__unable_to_get_token(
    pyperclip_patch, webbrowser_patch, pca_mock
):
    pca_mock.get_accounts.return_value = None
    pca_mock.initiate_device_flow.return_value = {
        "message": "TEST MESSAGE",
        "verification_uri": "TEST URL",
        "user_code": "TEST CODE",
    }
    pca_mock.acquire_token_by_device_flow.return_value = {
        "error": "BAD REQUEST",
        "error_description": "Request to get token was bad.",
    }
    request_mock = MagicMock()
    request_mock.headers = {}
    with pytest.raises(
        AuthenticationError,
        match=(
            r"Unable to get token\. Error: BAD REQUEST "
            r"\(Details: Request to get token was bad\.\)\."
        ),
    ):
        DeviceCodeAuth(client=pca_mock, scopes=["TEST SCOPE"])(request_mock)
    pca_mock.acquire_token_silent.assert_not_called()
    pca_mock.initiate_device_flow.assert_called_with(scopes=["TEST SCOPE"])
    pca_mock.acquire_token_by_device_flow.assert_called_with(
        {
            "message": "TEST MESSAGE",
            "verification_uri": "TEST URL",
            "user_code": "TEST CODE",
        }
    )
    webbrowser_patch.open.assert_called_with("TEST URL")
    pyperclip_patch.copy.assert_called_with("TEST CODE")


@patch.dict(os.environ, {}, clear=True)
@patch("msal.PublicClientApplication", autospec=True)
@patch("msal_requests_auth.auth.device_code.webbrowser")
@patch("msal_requests_auth.auth.device_code.pyperclip")
def test_device_code_auth__invalid_accounts(
    pyperclip_patch, webbrowser_patch, pca_mock
):
    pca_mock.get_accounts.return_value = [{"account": "TEST ACCOUNT"}]
    pca_mock.acquire_token_silent.return_value = None
    pca_mock.initiate_device_flow.return_value = {
        "message": "TEST MESSAGE",
        "verification_uri": "TEST URL",
        "user_code": "TEST CODE",
    }
    pca_mock.acquire_token_by_device_flow.return_value = {
        "token_type": "Bearer",
        "access_token": "TEST TOKEN",
    }
    request_mock = MagicMock()
    request_mock.headers = {}
    returned_request = DeviceCodeAuth(client=pca_mock, scopes=["TEST SCOPE"])(
        request_mock
    )
    pca_mock.acquire_token_silent.assert_called_with(
        scopes=["TEST SCOPE"], account={"account": "TEST ACCOUNT"}
    )
    pca_mock.initiate_device_flow.assert_called_with(scopes=["TEST SCOPE"])
    pca_mock.acquire_token_by_device_flow.assert_called_with(
        {
            "message": "TEST MESSAGE",
            "verification_uri": "TEST URL",
            "user_code": "TEST CODE",
        }
    )

    assert returned_request.headers == {"Authorization": "Bearer TEST TOKEN"}

    webbrowser_patch.open.assert_called_with("TEST URL")
    pyperclip_patch.copy.assert_called_with("TEST CODE")


@patch.dict(os.environ, {}, clear=True)
@patch("msal.PublicClientApplication", autospec=True)
@patch("msal_requests_auth.auth.device_code.webbrowser")
@patch("msal_requests_auth.auth.device_code.pyperclip")
def test_device_code_auth__accounts(pyperclip_patch, webbrowser_patch, pca_mock):
    pca_mock.get_accounts.return_value = [{"account": "TEST ACCOUNT"}]
    pca_mock.acquire_token_silent.return_value = {
        "token_type": "Bearer",
        "access_token": "TEST TOKEN",
        "user_code": "TEST CODE",
    }
    request_mock = MagicMock()
    request_mock.headers = {}
    returned_request = DeviceCodeAuth(client=pca_mock, scopes=["TEST SCOPE"])(
        request_mock
    )
    pca_mock.acquire_token_silent.assert_called_with(
        scopes=["TEST SCOPE"], account={"account": "TEST ACCOUNT"}
    )
    pca_mock.initiate_device_flow.assert_not_called()
    pca_mock.acquire_token_by_device_flow.assert_not_called()

    assert returned_request.headers == {"Authorization": "Bearer TEST TOKEN"}
    webbrowser_patch.open.assert_not_called()
    pyperclip_patch.copy.assert_not_called()
